open Ast

(* [is_value e] is whether [e] is a syntactic value *)
let is_value : expr -> bool = function
  | Int _ -> true
  | Add _ | Mult _ | Var _ | Let _ -> false

(* [subst e v x] is e{v/x}, that is, [e] with [v]
 * substituted for [x]. *)
let rec subst e v x = match e with
  | Var y       -> if x=y then v else e
  | Int n       -> Int n
  | Add(el,er)  -> Add(subst el v x, subst er v x)
  | Mult(el,er) -> Mult(subst el v x, subst er v x)  (* NEW *)
  | Let(y,ebind,ebody) ->
      let ebind' = subst ebind v x in
      if x=y
      then Let(y, ebind', ebody)
      else Let(y, ebind', subst ebody v x)

(* A single step of evaluation. *)
let rec step = function
  | Int n -> failwith "Does not step"
  | Var _ -> failwith "Unbound variable"
  | Add(Int n1, Int n2) -> Int (n1+n2)
  | Add(Int n1, e2) -> Add(Int n1, step e2)
  | Add(e1,e2) -> Add(step e1, e2)
  | Mult(Int n1, Int n2) -> Int (n1*n2)  (* NEW *)
  | Mult(Int n1, e2) -> Mult(Int n1, step e2)  (* NEW *)
  | Mult(e1,e2) -> Mult(step e1, e2)  (* NEW *)
  | Let(x,Int n,e2) -> subst e2 (Int n) x
  | Let(x,e1,e2) -> Let(x,step e1, e2)

(* [eval e] is the [e -->* v] judgement.  That is,
 * keep applying [step] until a value is produced.  *)
let rec eval : expr -> expr = fun e ->
  if is_value e then e
  else eval (step e)

(* Parse a string into an ast *)
let parse s =
  let lexbuf = Lexing.from_string s in
  let ast = Parser.prog Lexer.read lexbuf in
  ast

(* Extract a value from an ast node.
   Raises Failure if the argument is a node containing a value. *)
let extract_value = function
  | Int i -> i
  | _ -> failwith "Not a value"

(* Interpret an expression *)
let interp e =
  e |> parse |> eval |> extract_value






